//------------------------------------------------------------------------
//
// Author      : Maarten "McClaw" Kronberger
// Email       : sulacomcclaw@hotmail.com
// Website     : http://www.sulaco.co.za
// Date        : 01 June 2004
// Version     : 1.0
// Description : This tutorial shows how to use the GL_REFLECTION_MAP_EXT
//               and GL_TEXTURE_CUBE_MAP extensions to render a reflective Cube map
//
//               The dglOpenGL.pas file was created by the team at www.delphigl.com
//               I have included the html readme for an explanation of using this.
//
//               If you have any queries or bug reports, please mail me.
//------------------------------------------------------------------------
// Modifed     : Vereshagin Roman Vladimirovich
// email       : support@sibvrv.com
// web         : http://www.sibvrv.com
//------------------------------------------------------------------------

program OpenGLApp;

uses
  Windows,
  Messages,
  dglOpenGL,
  MetaBalls,
  Textures;
var
  SCREEN_WIDTH  : Integer = 800;
  SCREEN_HEIGHT : Integer = 600;
const
  WND_TITLE = 'Cube Mapping by McCLaw; Dynamic Cube Mapping by SibVRV';
  FPS_TIMER = 1;                     // Timer to calculate FPS
  FPS_INTERVAL = 1000;               // Calculate FPS every 1000 ms
  SCREEN_FULLSCREEN = false;
  SCREEN_PIXEL_DEPTH = 32;

  CUBE_MAP_SIZE  = 16;
  CubeMap_Define : array[0..5] of TGLenum = (
                                             GL_TEXTURE_CUBE_MAP_POSITIVE_X_ARB,
                                             GL_TEXTURE_CUBE_MAP_NEGATIVE_X_ARB,
                                             GL_TEXTURE_CUBE_MAP_POSITIVE_Y_ARB,
                                             GL_TEXTURE_CUBE_MAP_NEGATIVE_Y_ARB,
                                             GL_TEXTURE_CUBE_MAP_POSITIVE_Z_ARB,
                                             GL_TEXTURE_CUBE_MAP_NEGATIVE_Z_ARB
                                             );
var
  h_Wnd  : HWND;                     // Global window handle
  h_DC   : HDC;                      // Global device context
  h_RC   : HGLRC;                    // OpenGL rendering context
  keys : Array[0..255] of Boolean;   // Holds keystrokes
  FPSCount : Integer = 0;            // Counter for FPS
  ElapsedTime : Integer;             // Elapsed time between frames

  // Textures
  CubeMap_TEX     : GluInt;
  Texture_ball    : GLuint;
  EnvironmentBox  : array [0..5] of GLuint;

  // User vaiables
  ySpeed : glFloat;
  yAngle : glFloat;

  // Mouse Variables
  MouseMove : boolean = false;        // Is the mouse moving?
  MoveXcoord, MoveYcoord : GLfloat;   // Current X and Y position of the cursor

  MouseButton : Integer = -1;         // mouse button down
  Xcoord, Ycoord : Integer;           // Current X and Y position where the mouse was clicked

  OpenGLInitialized : Boolean;

  // Quadric for sphere
  SphereQuadratic : PgluQuadricObj;

{$R *.RES}

{------------------------------------------------------------------}
{  Function to convert int to string. (No sysutils = smaller EXE)  }
{------------------------------------------------------------------}
function IntToStr(Num : Integer) : String;  // using SysUtils increase file size by 100K
begin
  Str(Num, result);
end;

{------------------------------------------------------------------}
{  Handle window resize                                            }
{------------------------------------------------------------------}
procedure glResizeWnd(Width, Height : Integer);
begin
  if (Height = 0) then                // prevent divide by zero exception
    Height := 1;
  SCREEN_WIDTH  := Width;
  SCREEN_HEIGHT := Height;

  glViewport(0, 0, Width, Height);    // Set the viewport for the OpenGL window
  glMatrixMode(GL_PROJECTION);        // Change Matrix Mode to Projection
  glLoadIdentity();                   // Reset View
  gluPerspective(45.0, Width/Height, 1.0, 200.0);  // Do the perspective calculations. Last value = max clipping depth

  glMatrixMode(GL_MODELVIEW);         // Return to the modelview matrix
  glLoadIdentity();                   // Reset View
end;

{------------------------------------------------------------------}
{  Function to render the scene                                    }
{------------------------------------------------------------------}
procedure RenderScene();
var
  iBoxSize : Single;
begin
  iBoxSize :=8.0;
  // Bind The background Texture
  glPushMatrix;
    glRotatef(yAngle,0,1,0);
    //glRotatef(yAngle,1,0,0);
    glBindTexture(GL_TEXTURE_2D,EnvironmentBox[2]);
    glBegin(GL_QUADS);
      // Front Face
      glNormal3f( 0.0, 0.0, -1.0);
      glTexCoord2f(1.0, 1.0); glVertex3f(-iBoxSize, -iBoxSize,  iBoxSize);
      glTexCoord2f(0.0, 1.0); glVertex3f( iBoxSize, -iBoxSize,  iBoxSize);
      glTexCoord2f(0.0, 0.0); glVertex3f( iBoxSize,  iBoxSize,  iBoxSize);
      glTexCoord2f(1.0, 0.0); glVertex3f(-iBoxSize,  iBoxSize,  iBoxSize);
    glEnd;
    glBindTexture(GL_TEXTURE_2D,EnvironmentBox[3]);
    glBegin(GL_QUADS);
      // Back Face
      glNormal3f( 0.0, 0.0, 1.0);
      glTexCoord2f(1.0, 0.0); glVertex3f( iBoxSize,  iBoxSize, -iBoxSize);
      glTexCoord2f(1.0, 1.0); glVertex3f( iBoxSize, -iBoxSize, -iBoxSize);
      glTexCoord2f(0.0, 1.0); glVertex3f(-iBoxSize, -iBoxSize, -iBoxSize);
      glTexCoord2f(0.0, 0.0); glVertex3f(-iBoxSize,  iBoxSize, -iBoxSize);
    glEnd;
    glBindTexture(GL_TEXTURE_2D,EnvironmentBox[1]);
    glBegin(GL_QUADS);
      // Top Face
      glNormal3f( 0.0, -1.0, 0.0);
      glTexCoord2f(0.0, 1.0); glVertex3f(-iBoxSize,  iBoxSize, -iBoxSize);
      glTexCoord2f(0.0, 0.0); glVertex3f(-iBoxSize,  iBoxSize,  iBoxSize);
      glTexCoord2f(1.0, 0.0); glVertex3f( iBoxSize,  iBoxSize,  iBoxSize);
      glTexCoord2f(1.0, 1.0); glVertex3f( iBoxSize,  iBoxSize, -iBoxSize);
    glEnd;
    glBindTexture(GL_TEXTURE_2D,EnvironmentBox[0]);
    glBegin(GL_QUADS);
      // Bottom Face
      glNormal3f( 0.0, 1.0, 0.0);
      glTexCoord2f(1.0, 1.0); glVertex3f( iBoxSize, -iBoxSize,  iBoxSize);
      glTexCoord2f(0.0, 1.0); glVertex3f(-iBoxSize, -iBoxSize,  iBoxSize);
      glTexCoord2f(0.0, 0.0); glVertex3f(-iBoxSize, -iBoxSize, -iBoxSize);
      glTexCoord2f(1.0, 0.0); glVertex3f( iBoxSize, -iBoxSize, -iBoxSize);
    glEnd;
    glBindTexture(GL_TEXTURE_2D,EnvironmentBox[5]);
    glBegin(GL_QUADS);
      // Right face
      glNormal3f(-1.0, 0.0, 0.0);
      glTexCoord2f(1.0, 0.0); glVertex3f( iBoxSize,  iBoxSize,  iBoxSize);
      glTexCoord2f(1.0, 1.0); glVertex3f( iBoxSize, -iBoxSize,  iBoxSize);
      glTexCoord2f(0.0, 1.0); glVertex3f( iBoxSize, -iBoxSize, -iBoxSize);
      glTexCoord2f(0.0, 0.0); glVertex3f( iBoxSize,  iBoxSize, -iBoxSize);
    glEnd;
    glBindTexture(GL_TEXTURE_2D,EnvironmentBox[4]);
    glBegin(GL_QUADS);
      // Left Face
      glNormal3f( 1.0, 0.0, 0.0);
      glTexCoord2f(0.0, 0.0); glVertex3f(-iBoxSize,  iBoxSize,  iBoxSize);
      glTexCoord2f(1.0, 0.0); glVertex3f(-iBoxSize,  iBoxSize, -iBoxSize);
      glTexCoord2f(1.0, 1.0); glVertex3f(-iBoxSize, -iBoxSize, -iBoxSize);
      glTexCoord2f(0.0, 1.0); glVertex3f(-iBoxSize, -iBoxSize,  iBoxSize);
    glEnd;

    glBindTexture(GL_TEXTURE_2D,Texture_ball);
    glTranslatef(Sin(ElapsedTime/1500)*2,Cos(ElapsedTime/1500)*2,0);
    gluSphere(SphereQuadratic, 0.6, 32, 32);  // draw sphere
  glPopMatrix;
end;

{------------------------------------------------------------------}
{  Function to generate the cibemats for this frame                }
{------------------------------------------------------------------}
procedure CreateCubeMaps(PosX,PosY,PosZ : single);
var
 i:Integer;
begin
	glViewport(0,0,CUBE_MAP_SIZE,CUBE_MAP_SIZE);
	glScissor(0,0,CUBE_MAP_SIZE,CUBE_MAP_SIZE);
	glEnable(GL_SCISSOR_TEST);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(90,1,1,2000);
	glMatrixMode(GL_MODELVIEW);
  for i := 0 to 5 do begin
    glClear(GL_COLOR_BUFFER_BIT or GL_DEPTH_BUFFER_BIT);
    glLoadIdentity();
    case i of
    0:
      begin
        glRotatef( 90,0,1,0);
        glRotatef(180,1,0,0);
      end;
    1:
      begin
        glRotatef(-90,0,1,0);
        glRotatef(180,1,0,0);
      end;
    2: glRotatef(-90,1,0,0);
    3: glRotatef( 90,1,0,0);
    4: glRotatef(180,1,0,0);
    5: glRotatef(180,0,0,1);
    end;
    //
    glTranslatef(PosX,PosY,PosZ);
    glScalef(-1,-1,-1);
    RenderScene;

    glEnable(GL_TEXTURE_CUBE_MAP_ARB);
    glBindTexture(GL_TEXTURE_CUBE_MAP_ARB,CubeMap_TEX);
    glCopyTexSubImage2D(CubeMap_define[i],0,0,0,0,0,CUBE_MAP_SIZE,CUBE_MAP_SIZE);
    glDisable(GL_TEXTURE_CUBE_MAP_ARB);
  end;
	glScissor(0,0,SCREEN_WIDTH,SCREEN_HEIGHT);
	glDisable(GL_SCISSOR_TEST);
  glResizeWnd(SCREEN_WIDTH,SCREEN_HEIGHT);
end;

{------------------------------------------------------------------}
{  Function to draw the actual scene                               }
{------------------------------------------------------------------}
procedure glDraw();
begin
  CreateCubeMaps(0,0,-5);
  glClear(GL_COLOR_BUFFER_BIT or GL_DEPTH_BUFFER_BIT);    // Clear The Screen And The Depth Buffer
  glLoadIdentity();                                       // Reset The View

  glTranslatef(0.0,0.0,-5);

  glPushMatrix;

    // Assign The CubeMaps to the mesh defined below
    glEnable(GL_TEXTURE_CUBE_MAP_EXT);   // Start cube mapping
    glEnable(GL_TEXTURE_GEN_S);
    glEnable(GL_TEXTURE_GEN_T);
    glEnable(GL_TEXTURE_GEN_R);
    glBindTexture(GL_TEXTURE_CUBE_MAP_ARB,CubeMap_TEX);

    drawMetaballs(ElapsedTime);
    // Restore OpenGL States
    glBindTexture(GL_TEXTURE_CUBE_MAP_ARB,0);
    glDisable(GL_TEXTURE_CUBE_MAP_EXT);     // Stop Cube Mapping
    glDisable(GL_TEXTURE_GEN_S);
    glDisable(GL_TEXTURE_GEN_T);
    glDisable(GL_TEXTURE_GEN_R);


  glPopMatrix;

  RenderScene();
  yAngle := yAngle + ySpeed;

  // Flush the OpenGL Buffer
  glFlush();                            // ( Force the buffer to draw or send a network packet of commands in a networked system)
  glFinish();
end;


{------------------------------------------------------------------}
{  Initialise OpenGL                                               }
{------------------------------------------------------------------}
procedure glInit();
var
 i:Integer;
begin
  glClearColor(0.0, 0.0, 0.0, 0.0); 	   // Black Background
  glShadeModel(GL_SMOOTH);               // Enables Smooth Color Shading
  glClearDepth(1.0);                     // Depth Buffer Setup
  glEnable(GL_DEPTH_TEST);               // Enable Depth Buffer
  glDepthFunc(GL_LESS);		               // The Type Of Depth Test To Do

  glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);   //Realy Nice perspective calculations

  //glCullFace(GL_BACK);
  //glEnable(GL_CULL_FACE);

  glEnable(GL_TEXTURE_2D);               // Enable Texture Mapping

  // Load the background texture
  LoadTexture('yneg.jpg',EnvironmentBox[0]);
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
  LoadTexture('ypos.jpg',EnvironmentBox[1]);
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
  LoadTexture('zneg.jpg',EnvironmentBox[2]);
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
  LoadTexture('zpos.jpg',EnvironmentBox[3]);
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
  LoadTexture('xneg.jpg',EnvironmentBox[4]);
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
  LoadTexture('xpos.jpg',EnvironmentBox[5]);
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
  LoadTexture('ball.jpg',Texture_ball, false);


  // Enable S T and R Texture Coordinate Generation as a reflection map
  glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_REFLECTION_MAP_EXT);
  glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_REFLECTION_MAP_EXT);
  glTexGeni(GL_R, GL_TEXTURE_GEN_MODE, GL_REFLECTION_MAP_EXT);

  // Clamp all Cube Map Textures
  if GL_VERSION_1_2 then  // Is this card at least OpenGL ver 1.2 compliant?
  begin
    glTexParameteri(GL_TEXTURE_CUBE_MAP_EXT, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_CUBE_MAP_EXT, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
  end
  else
  begin
    glTexParameteri(GL_TEXTURE_CUBE_MAP_EXT, GL_TEXTURE_WRAP_S, GL_CLAMP);
    glTexParameteri(GL_TEXTURE_CUBE_MAP_EXT, GL_TEXTURE_WRAP_T, GL_CLAMP);
  end;

  glEnable(GL_TEXTURE_CUBE_MAP_EXT);
	glGenTextures(1,@CubeMap_TEX);
	glBindTexture(GL_TEXTURE_CUBE_MAP_ARB,CubeMap_TEX);
	glTexParameteri(GL_TEXTURE_CUBE_MAP_ARB,GL_TEXTURE_MAG_FILTER,
		GL_LINEAR);
	glTexParameteri(GL_TEXTURE_CUBE_MAP_ARB,GL_TEXTURE_MIN_FILTER,
		GL_LINEAR_MIPMAP_LINEAR);
	glTexParameteri(GL_TEXTURE_CUBE_MAP_ARB,GL_TEXTURE_WRAP_S,GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_CUBE_MAP_ARB,GL_TEXTURE_WRAP_T,GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_CUBE_MAP_ARB,GL_TEXTURE_WRAP_R,GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_CUBE_MAP_ARB,GL_GENERATE_MIPMAP_SGIS,GL_TRUE);
  for i := 0 to  5 do
  begin
      glTexImage2D(CubeMap_Define[i],0,GL_RGBA,CUBE_MAP_SIZE,CUBE_MAP_SIZE,0,GL_RGBA, GL_UNSIGNED_BYTE,0);
	end;
  glDisable(GL_TEXTURE_CUBE_MAP_ARB);
  
  InitMetaballs;

  // Generate a sphere using the OpenGL utility Library
  SphereQuadratic := gluNewQuadric();		       // Create A Pointer To The Quadric Object (Return 0 If No Memory) (NEW)
  gluQuadricNormals(SphereQuadratic, GLU_SMOOTH);	       // Create Smooth Normals
  gluQuadricTexture(SphereQuadratic, True);
  ySpeed := 0.1;   // start with some movement
end;


{------------------------------------------------------------------}
{  Processes all the mouse move                                  }
{------------------------------------------------------------------}
procedure ProcessMouseMove();
begin
  if MouseMove then
  begin
    // Add Code here to execute on mouse move
    MouseMove := False;
  end;
end;

{------------------------------------------------------------------}
{  Processes all the mouse clicks                                  }
{------------------------------------------------------------------}
procedure ProcessMouseClick;
begin
  case MouseButton of
  1: // Left Mouse Button
    begin
      // Add code to execute on mouse click
      MouseButton := 0; // Cancel our mouse click (To use this procedure as a mouse down event remove this line)
    end;
  2: // Right Mouse Button
    begin
      // Add code to execute on mouse click
      MouseButton := 0; // Cancel our mouse click (To use this procedure as a mouse down event remove this line)
    end;
  3: // Middle Mouse Button
    begin
      // Add code to execute on mouse click
      MouseButton := 0;  // Cancel our mouse click (To use this procedure as a mouse down event remove this line)
    end;
  end;
end;

{------------------------------------------------------------------}
{  Processes all the keystrokes                                    }
{------------------------------------------------------------------}
procedure ProcessKeys;
begin
  if (keys[VK_RIGHT]) then ySpeed := ySpeed + 0.002;
  if (keys[VK_LEFT])  then ySpeed := ySpeed - 0.002;
end;


{------------------------------------------------------------------}
{  Determines the applications response to the messages received  }
{------------------------------------------------------------------}
function WndProc(hWnd: HWND; Msg: UINT;  wParam: WPARAM;  lParam: LPARAM): LRESULT; stdcall;
begin
  case (Msg) of
    WM_CREATE:
      begin
        // Insert stuff you want executed when the program starts
      end;
    WM_CLOSE:
      begin
        PostQuitMessage(0);
        Result := 0
      end;
    WM_KEYDOWN:       // Set the pressed key (wparam) to equal true so we can check if its pressed
      begin
        keys[wParam] := True;
        Result := 0;
      end;
    WM_KEYUP:         // Set the released key (wparam) to equal false so we can check if its pressed
      begin
        keys[wParam] := False;
        Result := 0;
      end;
    WM_MOUSEMOVE:
      begin
        MouseMove := True;
        MoveXcoord := LOWORD(lParam);
        MoveYcoord := HIWORD(lParam);
        Result := 0;
      end;
    WM_LBUTTONDOWN:
      begin
        ReleaseCapture();   // need them here, because if mouse moves off
        SetCapture(h_Wnd);  // window and returns, it needs to reset status
        MouseButton := 1;
        Xcoord := LOWORD(lParam);
        Ycoord := HIWORD(lParam);
        Result := 0;
      end;
    WM_RBUTTONDOWN:
      begin
        ReleaseCapture();   // need them here, because if mouse moves off
        SetCapture(h_Wnd);  // window and returns, it needs to reset status
        MouseButton := 2;
        Xcoord := LOWORD(lParam);
        Ycoord := HIWORD(lParam);
        Result := 0;
      end;
    WM_MBUTTONDOWN:
      begin
        ReleaseCapture();   // need them here, because if mouse moves off
        SetCapture(h_Wnd);  // window and returns, it needs to reset status
        MouseButton := 3;
        Xcoord := LOWORD(lParam);
        Ycoord := HIWORD(lParam);
        Result := 0;
      end;
    WM_LBUTTONUP,WM_RBUTTONUP,WM_MBUTTONUP:
      begin
        ReleaseCapture();   // above
        MouseButton := 0;
        XCoord := 0;
        YCoord := 0;
        Result := 0;
      end;
    WM_SIZE:          // Resize the window with the new width and height
      begin
        if OpenGLInitialized then
         glResizeWnd(LOWORD(lParam),HIWORD(lParam));
        Result := 0;
      end;
    WM_TIMER :                     // Add code here for all timers to be used.
      begin
        if wParam = FPS_TIMER then
        begin
          FPSCount :=Round(FPSCount * 1000/FPS_INTERVAL);   // calculate to get per Second incase intercal is less or greater than 1 second
          SetWindowText(h_Wnd, PChar(WND_TITLE + '   [' + intToStr(FPSCount) + ' FPS]'));
          FPSCount := 0;
          Result := 0;
        end;
      end;
    else
      begin
      
      Result := DefWindowProc(hWnd, Msg, wParam, lParam);    // Default result if nothing happens
      end;
  end;
end;


{---------------------------------------------------------------------}
{  Properly destroys the window created at startup (no memory leaks)  }
{---------------------------------------------------------------------}
procedure glKillWnd(Fullscreen : Boolean);
begin
  if Fullscreen then             // Change back to non fullscreen
  begin
    ChangeDisplaySettings(devmode(nil^), 0);
    ShowCursor(True);
  end;

  // Makes current rendering context not current, and releases the device
  // context that is used by the rendering context.
  if (not wglMakeCurrent(h_DC, 0)) then
    MessageBox(0, 'Release of DC and RC failed!', 'Error', MB_OK or MB_ICONERROR);

  // Attempts to delete the rendering context
  if (not wglDeleteContext(h_RC)) then
  begin
    MessageBox(0, 'Release of rendering context failed!', 'Error', MB_OK or MB_ICONERROR);
    h_RC := 0;
  end;

  // Attemps to release the device context
  if ((h_DC > 0) and (ReleaseDC(h_Wnd, h_DC) = 0)) then
  begin
    MessageBox(0, 'Release of device context failed!', 'Error', MB_OK or MB_ICONERROR);
    h_DC := 0;
  end;

  // Attempts to destroy the window
  if ((h_Wnd <> 0) and (not DestroyWindow(h_Wnd))) then
  begin
    MessageBox(0, 'Unable to destroy window!', 'Error', MB_OK or MB_ICONERROR);
    h_Wnd := 0;
  end;

  // Attempts to unregister the window class
  if (not UnRegisterClass('OpenGL', hInstance)) then
  begin
    MessageBox(0, 'Unable to unregister window class!', 'Error', MB_OK or MB_ICONERROR);
    hInstance := 0;
  end;
end;


{--------------------------------------------------------------------}
{  Creates the window and attaches a OpenGL rendering context to it  }
{--------------------------------------------------------------------}
function glCreateWnd(Width, Height : Integer; Fullscreen : Boolean; PixelDepth : Integer) : Boolean;
var
  wndClass : TWndClass;         // Window class
  dwStyle : DWORD;              // Window styles
  dwExStyle : DWORD;            // Extended window styles
  dmScreenSettings : DEVMODE;   // Screen settings (fullscreen, etc...)
  PixelFormat : GLuint;         // Settings for the OpenGL rendering
  h_Instance : HINST;           // Current instance
  pfd : TPIXELFORMATDESCRIPTOR;  // Settings for the OpenGL window
begin
  InitOpenGL;                               // New call to initialize and bind the OpenGL dll

  h_Instance := GetModuleHandle(nil);       //Grab An Instance For Our Window
  ZeroMemory(@wndClass, SizeOf(wndClass));  // Clear the window class structure

  with wndClass do                    // Set up the window class
  begin
    style         := CS_HREDRAW or    // Redraws entire window if length changes
                     CS_VREDRAW or    // Redraws entire window if height changes
                     CS_OWNDC;        // Unique device context for the window
    lpfnWndProc   := @WndProc;        // Set the window procedure to our func WndProc
    hInstance     := h_Instance;
    hCursor       := LoadCursor(0, IDC_ARROW);
    lpszClassName := 'OpenGL';
  end;

  if (RegisterClass(wndClass) = 0) then  // Attemp to register the window class
  begin
    MessageBox(0, 'Failed to register the window class!', 'Error', MB_OK or MB_ICONERROR);
    Result := False;
    Exit
  end;

  // Change to fullscreen if so desired
  if Fullscreen then
  begin
    ZeroMemory(@dmScreenSettings, SizeOf(dmScreenSettings));
    with dmScreenSettings do begin              // Set parameters for the screen setting
      dmSize       := SizeOf(dmScreenSettings);
      dmPelsWidth  := Width;                    // Window width
      dmPelsHeight := Height;                   // Window height
      dmBitsPerPel := PixelDepth;               // Window color depth
      dmFields     := DM_PELSWIDTH or DM_PELSHEIGHT or DM_BITSPERPEL;
    end;

    // Try to change screen mode to fullscreen
    if (ChangeDisplaySettings(dmScreenSettings, CDS_FULLSCREEN) = DISP_CHANGE_FAILED) then
    begin
      MessageBox(0, 'Unable to switch to fullscreen!', 'Error', MB_OK or MB_ICONERROR);
      Fullscreen := False;
    end;
  end;

  // If we are still in fullscreen then
  if (Fullscreen) then
  begin
    dwStyle := WS_POPUP or                // Creates a popup window
               WS_CLIPCHILDREN            // Doesn't draw within child windows
               or WS_CLIPSIBLINGS;        // Doesn't draw within sibling windows
    dwExStyle := WS_EX_APPWINDOW;         // Top level window
    ShowCursor(False);                    // Turn of the cursor (gets in the way)
  end
  else
  begin
    dwStyle := WS_OVERLAPPEDWINDOW or     // Creates an overlapping window
               WS_CLIPCHILDREN or         // Doesn't draw within child windows
               WS_CLIPSIBLINGS;           // Doesn't draw within sibling windows
    dwExStyle := WS_EX_APPWINDOW or       // Top level window
                 WS_EX_WINDOWEDGE;        // Border with a raised edge
  end;

  // Attempt to create the actual window
  h_Wnd := CreateWindowEx(dwExStyle,      // Extended window styles
                          'OpenGL',       // Class name
                          WND_TITLE,      // Window title (caption)
                          dwStyle,        // Window styles
                          0, 0,           // Window position
                          Width, Height,  // Size of window
                          0,              // No parent window
                          0,              // No menu
                          h_Instance,     // Instance
                          nil);           // Pass nothing to WM_CREATE

  if h_Wnd = 0 then
  begin
    glKillWnd(Fullscreen);                // Undo all the settings we've changed
    MessageBox(0, 'Unable to create window!', 'Error', MB_OK or MB_ICONERROR);
    Result := False;
    Exit;
  end;

  // Try to get a device context
  h_DC := GetDC(h_Wnd);
  if (h_DC = 0) then
  begin
    glKillWnd(Fullscreen);
    MessageBox(0, 'Unable to get a device context!', 'Error', MB_OK or MB_ICONERROR);
    Result := False;
    Exit;
  end;

  // Settings for the OpenGL window
  with pfd do
  begin
    nSize           := SizeOf(TPIXELFORMATDESCRIPTOR); // Size Of This Pixel Format Descriptor
    nVersion        := 1;                    // The version of this data structure
    dwFlags         := PFD_DRAW_TO_WINDOW    // Buffer supports drawing to window
                       or PFD_SUPPORT_OPENGL // Buffer supports OpenGL drawing
                       or PFD_DOUBLEBUFFER;  // Supports double buffering
    iPixelType      := PFD_TYPE_RGBA;        // RGBA color format
    cColorBits      := PixelDepth;           // OpenGL color depth
    cRedBits        := 0;                    // Number of red bitplanes
    cRedShift       := 0;                    // Shift count for red bitplanes
    cGreenBits      := 0;                    // Number of green bitplanes
    cGreenShift     := 0;                    // Shift count for green bitplanes
    cBlueBits       := 0;                    // Number of blue bitplanes
    cBlueShift      := 0;                    // Shift count for blue bitplanes
    cAlphaBits      := 0;                    // Not supported
    cAlphaShift     := 0;                    // Not supported
    cAccumBits      := 0;                    // No accumulation buffer
    cAccumRedBits   := 0;                    // Number of red bits in a-buffer
    cAccumGreenBits := 0;                    // Number of green bits in a-buffer
    cAccumBlueBits  := 0;                    // Number of blue bits in a-buffer
    cAccumAlphaBits := 0;                    // Number of alpha bits in a-buffer
    cDepthBits      := 16;                   // Specifies the depth of the depth buffer
    cStencilBits    := 0;                    // Turn off stencil buffer
    cAuxBuffers     := 0;                    // Not supported
    iLayerType      := PFD_MAIN_PLANE;       // Ignored
    bReserved       := 0;                    // Number of overlay and underlay planes
    dwLayerMask     := 0;                    // Ignored
    dwVisibleMask   := 0;                    // Transparent color of underlay plane
    dwDamageMask    := 0;                     // Ignored
  end;

  // Attempts to find the pixel format supported by a device context that is the best match to a given pixel format specification.
  PixelFormat := ChoosePixelFormat(h_DC, @pfd);
  if (PixelFormat = 0) then
  begin
    glKillWnd(Fullscreen);
    MessageBox(0, 'Unable to find a suitable pixel format', 'Error', MB_OK or MB_ICONERROR);
    Result := False;
    Exit;
  end;

  // Sets the specified device context's pixel format to the format specified by the PixelFormat.
  if (not SetPixelFormat(h_DC, PixelFormat, @pfd)) then
  begin
    glKillWnd(Fullscreen);
    MessageBox(0, 'Unable to set the pixel format', 'Error', MB_OK or MB_ICONERROR);
    Result := False;
    Exit;
  end;

  // Create a OpenGL rendering context
  h_RC := wglCreateContext(h_DC);
  if (h_RC = 0) then
  begin
    glKillWnd(Fullscreen);
    MessageBox(0, 'Unable to create an OpenGL rendering context', 'Error', MB_OK or MB_ICONERROR);
    Result := False;
    Exit;
  end;

  // Makes the specified OpenGL rendering context the calling thread's current rendering context
  if (not wglMakeCurrent(h_DC, h_RC)) then
  begin
    glKillWnd(Fullscreen);
    MessageBox(0, 'Unable to activate OpenGL rendering context', 'Error', MB_OK or MB_ICONERROR);
    Result := False;
    Exit;
  end;
  // Read And Assign Extentions
  ReadExtensions;
  ReadImplementationProperties;
  OpenGLInitialized := True;
  // Initializes the timer used to calculate the FPS
  SetTimer(h_Wnd, FPS_TIMER, FPS_INTERVAL, nil);

  // Settings to ensure that the window is the topmost window
  ShowWindow(h_Wnd, SW_SHOW);
  SetForegroundWindow(h_Wnd);
  SetFocus(h_Wnd);

  // Ensure the OpenGL window is resized properly
  glResizeWnd(Width, Height);
  glInit();

  Result := True;
end;


{--------------------------------------------------------------------}
{  Main message loop for the application                             }
{--------------------------------------------------------------------}
function WinMain(hInstance : HINST; hPrevInstance : HINST;
                 lpCmdLine : PChar; nCmdShow : Integer) : Integer; stdcall;
var
  msg : TMsg;
  finished : Boolean;
  DemoStart, LastTime : DWord;
begin
  finished := False;

  // Perform application initialization:
  if not glCreateWnd(SCREEN_WIDTH, SCREEN_HEIGHT, SCREEN_FULLSCREEN, SCREEN_PIXEL_DEPTH) then
  begin
    Result := 0;
    Exit;
  end;

  DemoStart := GetTickCount();            // Get Time when demo started

  // Main message loop:
  while not finished do
  begin
    if (PeekMessage(msg, 0, 0, 0, PM_REMOVE)) then // Check if there is a message for this window
    begin
      if (msg.message = WM_QUIT) then     // If WM_QUIT message received then we are done
        finished := True
      else
      begin                               // Else translate and dispatch the message to this window
  	TranslateMessage(msg);
        DispatchMessage(msg);
      end;
    end
    else
    begin
      Inc(FPSCount);                      // Increment FPS Counter

      LastTime :=ElapsedTime;
      ElapsedTime :=GetTickCount() - DemoStart;     // Calculate Elapsed Time
      ElapsedTime :=(LastTime + ElapsedTime) DIV 2; // Average it out for smoother movement

      glDraw();                           // Draw the scene
      SwapBuffers(h_DC);                  // Display the scene

      if (keys[VK_ESCAPE]) then           // If user pressed ESC then set finised TRUE
        finished := True
      else
      begin
        ProcessKeys;                      // Check for any other key Pressed
        ProcessMouseMove;                 // Check for mouse move
        ProcessMouseClick;                // Check for Mouse Click
      end;
    end;
  end;
  glKillWnd(FALSE);
  Result := msg.wParam;
end;


begin
  WinMain( hInstance, hPrevInst, CmdLine, CmdShow );
end.
